# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2010-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import re
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from django.conf import settings
from django.contrib.auth.models import Group, Permission, User
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.test.utils import override_settings
from openid.consumer.consumer import (
    CancelResponse,
    FailureResponse,
    SetupNeededResponse,
    SuccessResponse,
)
from openid.consumer.discover import OpenIDServiceEndpoint
from openid.extensions import pape
from openid.message import Message, OPENID2_NS

from django_openid_auth.auth import OpenIDBackend, get_user_group_model
from django_openid_auth.exceptions import (
    DuplicateUsernameViolation,
    MissingPhysicalMultiFactor,
    MissingUsernameViolation,
    RequiredAttributeNotReturned,
)
from django_openid_auth.models import UserOpenID
from django_openid_auth.signals import openid_duplicate_username
from django_openid_auth.teams import ns_uri as TEAMS_NS
from django_openid_auth.tests.helpers import override_session_serializer


SREG_NS = "http://openid.net/sreg/1.0"
AX_NS = "http://openid.net/srv/ax/1.0"
SERVER_URL = 'http://example.com'


def make_claimed_id(id_):
    return urljoin(SERVER_URL, id_)


class TestMessage(Message):
    """Convenience class to construct test OpenID messages and responses."""

    def __init__(self, openid_namespace=OPENID2_NS):
        super(TestMessage, self).__init__(openid_namespace=openid_namespace)
        endpoint = OpenIDServiceEndpoint()
        endpoint.claimed_id = make_claimed_id('some-id')
        endpoint.server_url = SERVER_URL
        self.endpoint = endpoint

    def set_ax_args(
            self,
            email="foo@example.com",
            first=None,
            fullname="Some User",
            last=None,
            nickname="someuser",
            schema="http://axschema.org/",
            verified=False):

        attributes = [
            ("nickname", schema + "namePerson/friendly", nickname),
            ("fullname", schema + "namePerson", fullname),
            ("email", schema + "contact/email", email),
            ("account_verified",
             "http://ns.login.ubuntu.com/2013/validation/account",
             "token_via_email" if verified else "no")
            ]
        if first:
            attributes.append(
                ("first", "http://axschema.org/namePerson/first", first))
        if last:
            attributes.append(
                ("last", "http://axschema.org/namePerson/last", last))

        self.setArg(AX_NS, "mode", "fetch_response")
        for (alias, uri, value) in attributes:
            self.setArg(AX_NS, "type.%s" % alias, uri)
            self.setArg(AX_NS, "value.%s" % alias, value)

    def set_pape_args(self, *auth_policies):
        self.setArg(pape.ns_uri, 'auth_policies', ' '.join(auth_policies))

    def _set_args(self, ns, **kwargs):
        for key, value in kwargs.items():
            if value is not None:
                self.setArg(ns, key, value)
            elif self.hasKey(ns, key):
                self.delArg(ns, key)

    def set_sreg_args(self, **kwargs):
        self._set_args(SREG_NS, **kwargs)

    def set_team_args(self, **kwargs):
        self._set_args(TEAMS_NS, **kwargs)

    def to_response(self):
        return SuccessResponse(
            self.endpoint, self, signed_fields=self.toPostArgs().keys())


@override_session_serializer
@override_settings(
    OPENID_USE_EMAIL_FOR_USERNAME=False,
    OPENID_LAUNCHPAD_TEAMS_REQUIRED=[],
    OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=False,
    OPENID_EMAIL_WHITELIST_REGEXP_LIST=[])
class OpenIDBackendTests(TestCase):

    def setUp(self):
        super(OpenIDBackendTests, self).setUp()
        self.backend = OpenIDBackend()
        self.message = TestMessage()

    def make_user_openid(
            self, user=None, claimed_id=make_claimed_id('existing_identity')):
        if user is None:
            user = User.objects.create_user(
                username='someuser', email='someuser@example.com',
                password='12345678')

        user_openid, created = UserOpenID.objects.get_or_create(
            user=user, claimed_id=claimed_id, display_id=claimed_id)
        return user_openid

    def _assert_account_verified(self, user, expected):
        permission = Permission.objects.get(codename='account_verified')
        perm_label = '%s.%s' % (permission.content_type.app_label,
                                permission.codename)
        # Always invalidate the per-request perm cache
        attrs = list(user.__dict__.keys())
        for attr in attrs:
            if attr.endswith('_perm_cache'):
                delattr(user, attr)

        self.assertEqual(user.has_perm(perm_label), expected)

    def assert_account_not_verified(self, user):
        self._assert_account_verified(user, False)

    def assert_account_verified(self, user):
        self._assert_account_verified(user, True)

    def assert_no_users_created(self, expected_count=0):
        current_count = User.objects.count()
        msg = 'New users found (expected: %i, current: %i)' % (
            expected_count, current_count)
        self.assertEqual(current_count, expected_count, msg)

    def test_extract_user_details_sreg(self):
        expected = {
            'nickname': 'someuser',
            'first_name': 'Some',
            'last_name': 'User',
            'email': 'foo@example.com',
            'account_verified': False,
        }
        data = {
            'nickname': expected['nickname'],
            'fullname': "%s %s" % (expected['first_name'],
                                   expected['last_name']),
            'email': expected['email'],
        }
        self.message.set_sreg_args(**data)

        details = self.backend._extract_user_details(
            self.message.to_response())
        self.assertEqual(details, expected)

    def test_extract_user_details_ax(self):
        self.message.set_ax_args(
            email="foo@example.com",
            fullname="Some User",
            nickname="someuser",
        )
        data = self.backend._extract_user_details(self.message.to_response())

        self.assertEqual(data, {"nickname": "someuser",
                                "first_name": "Some",
                                "last_name": "User",
                                "email": "foo@example.com",
                                "account_verified": False})

    def test_extract_user_details_ax_split_name(self):
        # Include fullname too to show that the split data takes
        # precedence.
        self.message.set_ax_args(
            fullname="Bad Data", first="Some", last="User")
        data = self.backend._extract_user_details(self.message.to_response())

        self.assertEqual(data, {"nickname": "someuser",
                                "first_name": "Some",
                                "last_name": "User",
                                "email": "foo@example.com",
                                "account_verified": False})

    def test_extract_user_details_ax_broken_myopenid(self):
        self.message.set_ax_args(
            schema="http://schema.openid.net/", fullname="Some User",
            nickname="someuser", email="foo@example.com")
        data = self.backend._extract_user_details(self.message.to_response())

        self.assertEqual(data, {"nickname": "someuser",
                                "first_name": "Some",
                                "last_name": "User",
                                "email": "foo@example.com",
                                "account_verified": False})

    def test_update_user_details_long_names(self):
        self.message.set_ax_args()
        user = User.objects.create_user(
            'someuser', 'someuser@example.com', password=None)
        user_openid, created = UserOpenID.objects.get_or_create(
            user=user,
            claimed_id='http://example.com/existing_identity',
            display_id='http://example.com/existing_identity')
        data = dict(
            first_name=u"Some56789012345678901234567890123",
            last_name=u"User56789012345678901234567890123",
            email=u"someotheruser@example.com", account_verified=False)

        self.backend.update_user_details(
            user, data, self.message.to_response())

        self.assertEqual("Some56789012345678901234567890",  user.first_name)
        self.assertEqual("User56789012345678901234567890",  user.last_name)

    def _test_update_user_perms_account_verified(
            self, user, initially_verified, verified):
        # set user's verification status
        permission = Permission.objects.get(codename='account_verified')
        if initially_verified:
            user.user_permissions.add(permission)
        else:
            user.user_permissions.remove(permission)

        if initially_verified:
            self.assert_account_verified(user)
        else:
            self.assert_account_not_verified(user)

        # get a response including verification status
        self.message.set_ax_args()
        data = dict(first_name=u"Some56789012345678901234567890123",
                    last_name=u"User56789012345678901234567890123",
                    email=u"someotheruser@example.com",
                    account_verified=verified)
        self.backend.update_user_details(
            user, data, self.message.to_response())

        # refresh object from the database
        user = User.objects.get(pk=user.pk)

        if verified:
            self.assert_account_verified(user)
        else:
            self.assert_account_not_verified(user)

    def test_update_user_perms_initially_verified_then_verified(self):
        self._test_update_user_perms_account_verified(
            self.make_user_openid().user,
            initially_verified=True, verified=True)

    def test_update_user_perms_initially_verified_then_unverified(self):
        self._test_update_user_perms_account_verified(
            self.make_user_openid().user,
            initially_verified=True, verified=False)

    def test_update_user_perms_initially_not_verified_then_verified(self):
        self._test_update_user_perms_account_verified(
            self.make_user_openid().user,
            initially_verified=False, verified=True)

    def test_update_user_perms_initially_not_verified_then_unverified(self):
        self._test_update_user_perms_account_verified(
            self.make_user_openid().user,
            initially_verified=False, verified=False)

    def test_extract_user_details_name_with_trailing_space(self):
        self.message.set_ax_args(fullname="SomeUser ")

        data = self.backend._extract_user_details(self.message.to_response())

        self.assertEqual("", data['first_name'])
        self.assertEqual("SomeUser", data['last_name'])

    def test_extract_user_details_name_with_thin_space(self):
        self.message.set_ax_args(fullname=u"Some\u2009User")

        data = self.backend._extract_user_details(self.message.to_response())

        self.assertEqual("Some", data['first_name'])
        self.assertEqual("User", data['last_name'])

    @override_settings(OPENID_CREATE_USERS=True)
    def test_auth_username_when_no_nickname(self):
        self.message.set_sreg_args(nickname='')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)
        self.assertEqual(
            user.username, 'openiduser',
            "username must default to 'openiduser'")

    @override_settings(OPENID_USE_EMAIL_FOR_USERNAME=True)
    def test_auth_username_email_munging(self):
        for nick, email, expected in [
                ('nickcomesfirst', 'foo@example.com', 'nickcomesfirst'),
                ('', 'foo@example.com', 'fooexamplecom'),
                ('noemail', '', 'noemail'),
                ('', '@%.-', 'openiduser'),
                ('', '', 'openiduser'),
                (None, None, 'openiduser')]:
            self.message.set_sreg_args(nickname=nick, email=email)
            user = self.backend.authenticate(
                openid_response=self.message.to_response())
            # Cleanup user for further tests
            user.delete()

            self.assertIsNotNone(user)
            self.assertEqual(user.username, expected)

    @override_settings(OPENID_USE_EMAIL_FOR_USERNAME=False)
    def test_auth_username_no_email_munging(self):
        for nick, email, expected in [
                ('nickcomesfirst', 'foo@example.com', 'nickcomesfirst'),
                ('', 'foo@example.com', 'openiduser'),
                ('noemail', '', 'noemail'),
                ('', '@%.-', 'openiduser'),
                ('', '', 'openiduser'),
                (None, None, 'openiduser')]:
            self.message.set_sreg_args(nickname=nick, email=email)
            user = self.backend.authenticate(
                openid_response=self.message.to_response())
            # Cleanup user for further tests
            user.delete()

            self.assertIsNotNone(user)
            self.assertEqual(user.username, expected)

    @override_settings(
        OPENID_CREATE_USERS=True,
        OPENID_FOLLOW_RENAMES=False,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_username_duplicate_numbering(self):
        # Setup existing user to conflict with
        User.objects.create_user('testuser')

        self.message.set_sreg_args(nickname='testuser')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)
        self.assertEqual(
            user.username, 'testuser2',
            'Username must contain numeric suffix to avoid collisions.')

    def test_auth_username_duplicate_numbering_with_conflicts(self):
        # Setup existing users to conflict with
        User.objects.create_user('testuser')
        User.objects.create_user('testuser3')

        self.message.set_sreg_args(nickname='testuser')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        # Since this username is already taken by someone else, we go through
        # the process of adding +i to it starting with the count of users with
        # username starting with 'testuser', of which there are 2.  i should
        # start at 3, which already exists, so it should skip to 4.
        self.assertIsNotNone(user)
        self.assertEqual(
            user.username, 'testuser4',
            'Username must contain numeric suffix to avoid collisions.')

    def test_auth_username_duplicate_numbering_with_holes(self):
        # Setup existing users to conflict with
        User.objects.create_user('testuser')
        User.objects.create_user('testuser1')
        User.objects.create_user('testuser6')
        User.objects.create_user('testuser7')
        User.objects.create_user('testuser8')

        self.message.set_sreg_args(nickname='testuser')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        # Since this username is already taken by someone else, we go through
        # the process of adding +i to it starting with the count of users with
        # username starting with 'testuser', of which there are 5.  i should
        # start at 6, and increment until it reaches 9.
        self.assertIsNotNone(user)
        self.assertEqual(
            user.username, 'testuser9',
            'Username must contain numeric suffix to avoid collisions.')

    def test_auth_username_duplicate_numbering_with_nonsequential_matches(
            self):
        # Setup existing users to conflict with
        User.objects.create_user('testuser')
        User.objects.create_user('testuserfoo')

        self.message.set_sreg_args(nickname='testuser')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        # Since this username is already taken by someone else, we go through
        # the process of adding +i to it starting with the count of users with
        # username starting with 'testuser', of which there are 2.  i should
        # start at 3, which will be available.
        self.assertIsNotNone(user)
        self.assertEqual(
            user.username, 'testuser3',
            'Username must contain numeric suffix to avoid collisions.')

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team'])
    def test_authenticate_when_not_member_of_teams_required(self):
        Group.objects.create(name='team')

        self.message.set_sreg_args(nickname='someuser')
        self.message.set_team_args(is_member='foo')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNone(user)

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team'])
    def test_authenticate_when_no_group_mapping_to_required_team(self):
        assert Group.objects.filter(name='team').count() == 0

        self.message.set_sreg_args(nickname='someuser')
        self.message.set_team_args(is_member='foo')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNone(user)

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team'])
    def test_authenticate_when_member_of_teams_required(self):
        Group.objects.create(name='team')

        self.message.set_sreg_args(nickname='someuser')
        self.message.set_team_args(is_member='foo,team')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

    @override_settings(OPENID_LAUNCHPAD_TEAMS_REQUIRED=[])
    def test_authenticate_when_no_teams_required(self):
        self.message.set_sreg_args(nickname='someuser')
        self.message.set_team_args(is_member='team')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team1', 'team2'])
    def test_authenticate_when_member_of_at_least_one_team(self):
        Group.objects.create(name='team1')

        self.message.set_sreg_args(nickname='someuser')
        self.message.set_team_args(is_member='foo,team1')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team'],
        OPENID_EMAIL_WHITELIST_REGEXP_LIST=['foo(\+[^@]*)?@foo.com'])
    def test_authenticate_when_not_in_required_team_but_email_whitelisted(
            self):
        assert Group.objects.filter(name='team').count() == 0

        self.message.set_sreg_args(
            nickname='someuser', email='foo@foo.com')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

        self.message.set_sreg_args(
            nickname='someuser', email='foo+bar@foo.com')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team'],
        OPENID_EMAIL_WHITELIST_REGEXP_LIST=['foo@foo.com', 'bar@foo.com'])
    def test_authenticate_whitelisted_email_multiple_patterns(self):
        assert Group.objects.filter(name='team').count() == 0

        self.message.set_sreg_args(nickname='someuser', email='bar@foo.com')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

    @override_settings(
        OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO=True,
        OPENID_LAUNCHPAD_TEAMS_REQUIRED=['team'],
        OPENID_EMAIL_WHITELIST_REGEXP_LIST=['foo@foo.com'])
    def test_authenticate_whitelisted_email_not_match(self):
        assert Group.objects.filter(name='team').count() == 0

        self.message.set_sreg_args(nickname='someuser', email='bar@foo.com')
        self.message.set_team_args(is_member='foo')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNone(user)

    def test_auth_no_response(self):
        self.assertIsNone(self.backend.authenticate())
        self.assert_no_users_created()

    def test_auth_cancel_response(self):
        response = CancelResponse(OpenIDServiceEndpoint())

        self.assertIsNone(self.backend.authenticate(openid_response=response))
        self.assert_no_users_created()

    def test_auth_failure_response(self):
        response = FailureResponse(OpenIDServiceEndpoint())

        self.assertIsNone(self.backend.authenticate(openid_response=response))
        self.assert_no_users_created()

    def test_auth_setup_needed_response(self):
        response = SetupNeededResponse(OpenIDServiceEndpoint())

        self.assertIsNone(self.backend.authenticate(openid_response=response))
        self.assert_no_users_created()

    @override_settings(OPENID_CREATE_USERS=False)
    def test_auth_no_create_users(self):
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNone(user)
        self.assert_no_users_created()

    @override_settings(OPENID_CREATE_USERS=False)
    def test_auth_no_create_users_existing_user(self):
        existing_openid = self.make_user_openid(
            claimed_id=self.message.endpoint.claimed_id)
        expected_user_count = User.objects.count()
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)
        self.assertEqual(user, existing_openid.user)
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_UPDATE_DETAILS_FROM_SREG=True,
        OPENID_VALID_VERIFICATION_SCHEMES={
            SERVER_URL: {'token_via_email'}})
    def test_auth_update_details_from_sreg(self):
        first_name = 'a' * 31
        last_name = 'b' * 31
        email = 'new@email.com'
        self.message.set_ax_args(
            fullname=first_name + ' ' + last_name,
            nickname='newnickname',
            email=email,
            first=first_name,
            last=last_name,
            verified=True,
        )
        existing_openid = self.make_user_openid(
            claimed_id=self.message.endpoint.claimed_id)
        original_username = existing_openid.user.username
        expected_user_count = User.objects.count()

        self.assert_account_not_verified(existing_openid.user)

        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(user, existing_openid.user)
        self.assertEqual(
            user.username, original_username,
            'Username must not be updated unless OPENID_FOLLOW_RENAMES=True.')
        self.assertEqual(user.email, email)
        self.assertEqual(user.first_name, first_name[:30])
        self.assertEqual(user.last_name, last_name[:30])
        self.assert_account_verified(user)
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_UPDATE_DETAILS_FROM_SREG=True,
        OPENID_VALID_VERIFICATION_SCHEMES={
            SERVER_URL: {'token_via_email'}})
    def test_auth_update_details_from_sreg_unverifies_account(self):
        first_name = 'a' * 31
        last_name = 'b' * 31
        email = 'new@email.com'
        kwargs = dict(
            fullname=first_name + ' ' + last_name,
            nickname='newnickname',
            email=email,
            first=first_name,
            last=last_name,
            verified=True,
        )
        self.message.set_ax_args(**kwargs)
        verified_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assert_account_verified(verified_user)
        expected_user_count = User.objects.count()

        kwargs['verified'] = False
        self.message.set_ax_args(**kwargs)
        unverified_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(verified_user, unverified_user)
        self.assert_account_not_verified(unverified_user)
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(OPENID_PHYSICAL_MULTIFACTOR_REQUIRED=True)
    def test_physical_multifactor_required_not_given(self):
        response = self.message.to_response()

        with self.assertRaises(MissingPhysicalMultiFactor):
            self.backend.authenticate(openid_response=response)

        self.assertTrue(
            UserOpenID.objects.filter(
                claimed_id=self.message.endpoint.claimed_id).exists(),
            'User must be created anyways.')

    @override_settings(OPENID_PHYSICAL_MULTIFACTOR_REQUIRED=True)
    def test_physical_multifactor_required_invalid_auth_policy(self):
        self.message.set_pape_args(
            pape.AUTH_MULTI_FACTOR, pape.AUTH_PHISHING_RESISTANT)

        with self.assertRaises(MissingPhysicalMultiFactor):
            self.backend.authenticate(
                openid_response=self.message.to_response())

        self.assertTrue(
            UserOpenID.objects.filter(
                claimed_id=self.message.endpoint.claimed_id).exists(),
            'User must be created anyways.')

    @override_settings(OPENID_PHYSICAL_MULTIFACTOR_REQUIRED=True)
    def test_physical_multifactor_required_valid_auth_policy(self):
        self.message.set_pape_args(
            pape.AUTH_MULTI_FACTOR, pape.AUTH_MULTI_FACTOR_PHYSICAL,
            pape.AUTH_PHISHING_RESISTANT)
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)

    @override_settings(OPENID_STRICT_USERNAMES=True)
    def test_auth_strict_usernames(self):
        username = 'nickname'
        self.message.set_sreg_args(nickname=username)
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user, 'User must be created')
        self.assertEqual(user.username, username)

    @override_settings(OPENID_STRICT_USERNAMES=True)
    def test_auth_strict_usernames_no_nickname(self):
        self.message.set_sreg_args(nickname='')

        msg = re.escape(
            "An attribute required for logging in was not returned (nickname)")

        with self.assertRaisesRegexp(RequiredAttributeNotReturned, msg):
            self.backend.authenticate(
                openid_response=self.message.to_response())

        self.assert_no_users_created()

    @override_settings(
        OPENID_STRICT_USERNAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_strict_usernames_conflict(self):
        existing_openid = self.make_user_openid()
        expected_user_count = User.objects.count()
        self.message.set_sreg_args(
            nickname=existing_openid.user.username)

        with self.assertRaises(DuplicateUsernameViolation):
            self.backend.authenticate(
                openid_response=self.message.to_response())

        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames(self):
        new_username = 'new'
        self.message.set_sreg_args(nickname='username')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())
        expected_user_count = User.objects.count()

        self.assertIsNotNone(user, 'User must be created')

        self.message.set_sreg_args(nickname=new_username)
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(user.pk, renamed_user.pk)
        self.assertEqual(renamed_user.username, new_username)
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_STRICT_USERNAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_strict_usernames_no_nickname(self):
        self.message.set_sreg_args(nickname='nickame')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())
        expected_user_count = User.objects.count()

        self.assertIsNotNone(user, 'User must be created')

        self.message.set_sreg_args(nickname='')

        # XXX: Check possibilities to normalize this error into a
        # `RequiredAttributeNotReturned`.
        with self.assertRaises(MissingUsernameViolation):
            self.backend.authenticate(
                openid_response=self.message.to_response())

        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_STRICT_USERNAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_strict_usernames_rename_conflict(self):
        # Setup existing user to conflict with
        User.objects.create_user('testuser')

        self.message.set_sreg_args(nickname='nickname')
        user = self.backend.authenticate(
            openid_response=self.message.to_response())
        expected_user_count = User.objects.count()

        self.assertIsNotNone(user, 'First request should succeed')

        self.message.set_sreg_args(nickname='testuser')

        with self.assertRaises(DuplicateUsernameViolation):
            self.backend.authenticate(
                openid_response=self.message.to_response())

        db_user = User.objects.get(pk=user.pk)
        self.assertEqual(db_user.username, 'nickname')
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_STRICT_USERNAMES=False,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_to_conflict(self):
        # Setup existing user to conflict with
        User.objects.create_user('testuser')
        # Setup user to rename
        user = User.objects.create_user('nickname')
        self.make_user_openid(
            user=user, claimed_id=self.message.endpoint.claimed_id)
        # Trigger rename
        self.message.set_sreg_args(nickname='testuser')
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(renamed_user.pk, user.pk)
        self.assertEqual(
            renamed_user.username, 'testuser2',
            'Username must have a numbered suffix to avoid conflict.')

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_no_change(self):
        # Setup user to rename
        user = User.objects.create_user('username')
        self.make_user_openid(
            user=user, claimed_id=self.message.endpoint.claimed_id)
        expected_user_count = User.objects.count()
        # Trigger rename
        self.message.set_sreg_args(nickname=user.username)
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(renamed_user.pk, user.pk)
        self.assertEqual(
            renamed_user.username, user.username,
            'No numeric suffix should be appended for username owner.')
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_to_numbered_suffix(self):
        # Setup user to rename to numbered suffix pattern
        user = User.objects.create_user('testuser2000eight')
        self.make_user_openid(
            user=user, claimed_id=self.message.endpoint.claimed_id)
        # Trigger rename
        self.message.set_sreg_args(nickname='testuser2')
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(renamed_user.pk, user.pk)
        self.assertEqual(
            renamed_user.username, 'testuser2',
            'The numbered suffix must be kept.')

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_to_numbered_suffix_with_existing(self):
        # Setup existing user to conflict with
        User.objects.create_user('testuser')
        # Setup user to rename to numbered suffix pattern
        user = User.objects.create_user('testuser2000eight')
        self.make_user_openid(
            user=user, claimed_id=self.message.endpoint.claimed_id)
        # Trigger rename
        self.message.set_sreg_args(nickname='testuser3')
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(renamed_user.pk, user.pk)
        self.assertEqual(
            renamed_user.username, 'testuser3',
            'Username must be kept as there are no conflicts.')

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_from_numbered_suffix_to_conflict(self):
        # Setup existing user to conflict with
        User.objects.create_user('testuser')
        # Setup user with numbered suffix pattern
        user = User.objects.create_user('testuser2000')
        self.make_user_openid(
            user=user, claimed_id=self.message.endpoint.claimed_id)
        # Trigger rename
        self.message.set_sreg_args(nickname='testuser')
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(renamed_user.pk, user.pk)
        self.assertEqual(
            user.username, 'testuser2000',
            'Since testuser conflicts, username must remain unchanged as it '
            'maches the number suffix pattern.')

    @override_settings(
        OPENID_FOLLOW_RENAMES=True,
        OPENID_UPDATE_DETAILS_FROM_SREG=True)
    def test_auth_follow_renames_from_numbered_suffix_no_conflict(self):
        # Setup user with numbered suffix pattern
        user = User.objects.create_user('testuser2')
        self.make_user_openid(
            user=user, claimed_id=self.message.endpoint.claimed_id)
        # Trigger rename
        self.message.set_sreg_args(nickname='testuser')
        renamed_user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertEqual(renamed_user.pk, user.pk)
        self.assertEqual(
            renamed_user.username, 'testuser',
            'Username must be updated as there are no conflicts.')

    @override_settings(OPENID_STRICT_USERNAMES=True)
    def test_auth_duplicate_username_signal_is_sent(self):
        existing_openid = self.make_user_openid()
        expected_user_count = User.objects.count()
        signal_kwargs = {}

        def duplicate_username_handler(sender, **kwargs):
            signal_kwargs.update(kwargs)
        self.addCleanup(
            openid_duplicate_username.disconnect,
            duplicate_username_handler, sender=User, dispatch_uid='testing')
        openid_duplicate_username.connect(
            duplicate_username_handler, sender=User, weak=False,
            dispatch_uid='testing')

        self.message.set_sreg_args(
                nickname=existing_openid.user.username)

        with self.assertRaises(DuplicateUsernameViolation):
            self.backend.authenticate(
                openid_response=self.message.to_response())

        self.assertIn('username', signal_kwargs)
        self.assertEqual(
            signal_kwargs['username'], existing_openid.user.username)
        self.assert_no_users_created(expected_count=expected_user_count)

    @override_settings(OPENID_STRICT_USERNAMES=True)
    def test_auth_duplicate_username_signal_can_prevent_duplicate_error(self):
        existing_openid = self.make_user_openid()

        def duplicate_username_handler(sender, **kwargs):
            existing_user = existing_openid.user
            existing_user.username += '_other'
            existing_user.save()
        self.addCleanup(
            openid_duplicate_username.disconnect,
            duplicate_username_handler, sender=User, dispatch_uid='testing')
        openid_duplicate_username.connect(
            duplicate_username_handler, sender=User, weak=False,
            dispatch_uid='testing')

        self.message.set_sreg_args(
                nickname=existing_openid.user.username)
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)
        self.assertNotEqual(user, existing_openid.user)

    @override_settings(OPENID_STRICT_USERNAMES=True)
    def test_auth_duplicate_username_is_not_called_if_no_conflict(self):
        def duplicate_username_handler(sender, **kwargs):
            assert False, 'This should never have been called.'
        self.addCleanup(
            openid_duplicate_username.disconnect,
            duplicate_username_handler, sender=User, dispatch_uid='testing')
        openid_duplicate_username.connect(
            duplicate_username_handler, sender=User, weak=False,
            dispatch_uid='testing')

        self.message.set_sreg_args(nickname='nickname')
        self.backend.authenticate(openid_response=self.message.to_response())

    @override_settings(OPENID_STRICT_USERNAMES=False)
    def test_auth_duplicate_username_is_not_called_if_not_strict(self):
        existing_openid = self.make_user_openid()

        def duplicate_username_handler(sender, **kwargs):
            assert False, 'This should never have been called.'
        self.addCleanup(
            openid_duplicate_username.disconnect,
            duplicate_username_handler, sender=User, dispatch_uid='testing')
        openid_duplicate_username.connect(
            duplicate_username_handler, sender=User, weak=False,
            dispatch_uid='testing')

        self.message.set_sreg_args(nickname=existing_openid.user.username)
        self.backend.authenticate(openid_response=self.message.to_response())

    @override_settings(OPENID_STRICT_USERNAMES=True)
    def test_auth_duplicate_username_handling_bypass_numbered_suffix(self):
        nickname = 'nickname87'
        existing_openid = self.make_user_openid(
            user=User.objects.create_user(nickname))

        def duplicate_username_handler(sender, **kwargs):
            existing_user = existing_openid.user
            existing_user.username += '00'
            existing_user.save()
        self.addCleanup(
            openid_duplicate_username.disconnect,
            duplicate_username_handler, sender=User, dispatch_uid='testing')
        openid_duplicate_username.connect(
            duplicate_username_handler, sender=User, weak=False,
            dispatch_uid='testing')

        self.message.set_sreg_args(nickname=existing_openid.user.username)
        user = self.backend.authenticate(
            openid_response=self.message.to_response())

        self.assertIsNotNone(user)
        self.assertNotEqual(user, existing_openid.user)
        self.assertEqual(
            user.username, nickname,
            'In strict mode, when conflicts are handled, the username must '
            'be kept unmodified without numbered suffixes.')


class GetGroupModelTestCase(TestCase):

    def setUp(self):
        super(GetGroupModelTestCase, self).setUp()
        self.inject_test_models()

    def inject_test_models(self):
        installed_apps = settings.INSTALLED_APPS + (
            'django_openid_auth.tests',
        )
        p = self.settings(INSTALLED_APPS=installed_apps)
        p.enable()
        self.addCleanup(p.disable)
        self.clear_app_cache()

    def clear_app_cache(self):
        try:
            from django.apps import apps
            apps.clear_cache()
        except ImportError:
            from django.db.models.loading import cache
            cache.loaded = False

    def test_default_group_model(self):
        model = get_user_group_model()
        self.assertEqual(model, User.groups.through)

    @override_settings(AUTH_USER_GROUP_MODEL='tests.UserGroup')
    def test_custom_group_model(self):
        from django_openid_auth.tests.models import UserGroup
        model = get_user_group_model()
        self.assertEqual(model, UserGroup)

    @override_settings(
        AUTH_USER_GROUP_MODEL='django_openid_auth.models.UserGroup')
    def test_improperly_configured_invalid_name(self):
        self.assertRaises(ImproperlyConfigured, get_user_group_model)

    @override_settings(
        AUTH_USER_GROUP_MODEL='invalid.UserGroup')
    def test_improperly_configured_invalid_app(self):
        self.assertRaises(ImproperlyConfigured, get_user_group_model)



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2009-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Tests for the django_openid_auth Admin login form replacement."""

from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth.models import User

from django.test import TestCase
from django.test.utils import override_settings


@override_settings(OPENID_USE_AS_ADMIN_LOGIN=True)
class SiteAdminTests(TestCase):
    """
    TestCase for accessing /admin/ when the django_openid_auth form replacement
    is in use.
    """

    def test_admin_site_with_openid_login_authenticated_non_staff(self):
        """
        If the request has an authenticated user, who is not flagged as a
        staff member, then they get a failure response.
        """
        User.objects.create_user(
            username='testing', email='testing@example.com', password='test')
        assert self.client.login(username='testing', password='test')
        response = self.client.get('/admin/', follow=True)
        self.assertContains(
            response,
            'User testing does not have admin/staff access.', status_code=403)

    def test_admin_site_with_openid_login_non_authenticated_user(self):
        """
        Unauthenticated users accessing the admin page should be directed to
        the OpenID login url.
        """
        response = self.client.get('/admin/', follow=True)
        self.assertRedirects(
            response,
            getattr(settings, 'LOGIN_URL', '/openid/login') +
            '?next=%2Fadmin%2F')



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2007 Simon Willison
# Copyright (C) 2008-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.utils.html import escape


def index(request):
    s = ['<p>']
    if request.user.is_authenticated():
        s.append('You are signed in as <strong>%s</strong> (%s)' % (
                escape(request.user.username),
                escape(request.user.get_full_name())))
        s.append(' | <a href="/logout">Sign out</a>')
    else:
        s.append('<a href="/openid/login">Sign in with OpenID</a>')

    s.append('</p>')

    s.append('<p><a href="/private">This requires authentication</a></p>')
    return HttpResponse('\n'.join(s))


def next_works(request):
    return HttpResponse('?next= bit works. <a href="/">Home</a>')


@login_required
def require_authentication(request):
    return HttpResponse('This page requires authentication')



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2007 Simon Willison
# Copyright (C) 2008-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

from django import forms
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import UserChangeForm
from django.contrib.auth.models import Group
from django.utils.translation import ugettext as _
from django.conf import settings

from openid.yadis import xri

from django_openid_auth import PY3


def teams_new_unicode(self):
    """
    Replacement for Group.__unicode__()
    Calls original method to chain results
    """
    name = self.unicode_before_teams()
    teams_mapping = getattr(settings, 'OPENID_LAUNCHPAD_TEAMS_MAPPING', {})
    group_teams = [t for t in teams_mapping if teams_mapping[t] == self.name]
    if len(group_teams) > 0:
        return "%s -> %s" % (name, ", ".join(group_teams))
    else:
        return name


if PY3:
    Group.unicode_before_teams = Group.__str__
    Group.__str__ = teams_new_unicode
else:
    Group.unicode_before_teams = Group.__unicode__
    Group.__unicode__ = teams_new_unicode


class UserChangeFormWithTeamRestriction(UserChangeForm):
    """
    Extends UserChangeForm to add teams awareness to the user admin form
    """
    def clean_groups(self):
        data = self.cleaned_data['groups']
        teams_mapping = getattr(settings, 'OPENID_LAUNCHPAD_TEAMS_MAPPING', {})
        known_teams = teams_mapping.values()
        user_groups = self.instance.groups.all()
        for group in data:
            if group.name in known_teams and group not in user_groups:
                raise forms.ValidationError(
                    "The group %s is mapped to an external team.  "
                    "You cannot assign it manually." % group.name)
        return data


UserAdmin.form = UserChangeFormWithTeamRestriction


class OpenIDLoginForm(forms.Form):
    openid_identifier = forms.CharField(
        max_length=255,
        widget=forms.TextInput(attrs={'class': 'required openid'}))

    def clean_openid_identifier(self):
        if 'openid_identifier' in self.cleaned_data:
            openid_identifier = self.cleaned_data['openid_identifier']
            if (xri.identifierScheme(openid_identifier) == 'XRI' and
                    getattr(settings, 'OPENID_DISALLOW_INAMES', False)):
                raise forms.ValidationError(_('i-names are not supported'))
            return self.cleaned_data['openid_identifier']



#!/usr/bin/env python
# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2009-2015 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""OpenID integration for django.contrib.auth

A library that can be used to add OpenID support to Django applications.
The library integrates with Django's built in authentication system, so
most applications require minimal changes to support OpenID llogin. The
library also includes the following features:
  * Basic user details are transferred from the OpenID server via the
    Simple Registration extension or Attribute Exchange extension.
  * can be configured to use a fixed OpenID server URL, for use in SSO.
  * supports the launchpad.net teams extension to get team membership
    info.
"""

import sys

from setuptools import find_packages, setup

PY3 = sys.version_info.major >= 3


description, long_description = __doc__.split('\n\n', 1)
VERSION = '0.15'

install_requires = ['django>=1.6', 'six']
if PY3:
    install_requires.append('python3-openid')
else:
    install_requires.append('python-openid>=2.2.0')

setup(
    name='django-openid-auth',
    version=VERSION,

    packages=find_packages(),
    install_requires=install_requires,
    package_data={
        'django_openid_auth': ['templates/openid/*.html'],
    },

    # metadata for upload to PyPI
    author='Canonical Ltd',
    author_email='noreply@canonical.com',
    description=description,
    long_description=long_description,
    license='BSD',
    platforms=['any'],
    url='https://launchpad.net/django-openid-auth',
    download_url=('http://launchpad.net/django-openid-auth/trunk/%s/+download'
                  '/django-openid-auth-%s.tar.gz' % (VERSION, VERSION)),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],
)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Nonce'
        db.create_table(u'django_openid_auth_nonce', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('server_url', self.gf('django.db.models.fields.CharField')(max_length=2047)),
            ('timestamp', self.gf('django.db.models.fields.IntegerField')()),
            ('salt', self.gf('django.db.models.fields.CharField')(max_length=40)),
        ))
        db.send_create_signal(u'django_openid_auth', ['Nonce'])

        # Adding model 'Association'
        db.create_table(u'django_openid_auth_association', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('server_url', self.gf('django.db.models.fields.TextField')(max_length=2047)),
            ('handle', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('secret', self.gf('django.db.models.fields.TextField')(max_length=255)),
            ('issued', self.gf('django.db.models.fields.IntegerField')()),
            ('lifetime', self.gf('django.db.models.fields.IntegerField')()),
            ('assoc_type', self.gf('django.db.models.fields.TextField')(max_length=64)),
        ))
        db.send_create_signal(u'django_openid_auth', ['Association'])

        # Adding model 'UserOpenID'
        db.create_table(u'django_openid_auth_useropenid', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'])),
            ('claimed_id', self.gf('django.db.models.fields.TextField')(unique=True, max_length=2047)),
            ('display_id', self.gf('django.db.models.fields.TextField')(max_length=2047)),
        ))
        db.send_create_signal(u'django_openid_auth', ['UserOpenID'])


    def backwards(self, orm):
        # Deleting model 'Nonce'
        db.delete_table(u'django_openid_auth_nonce')

        # Deleting model 'Association'
        db.delete_table(u'django_openid_auth_association')

        # Deleting model 'UserOpenID'
        db.delete_table(u'django_openid_auth_useropenid')


    models = {
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'django_openid_auth.association': {
            'Meta': {'object_name': 'Association'},
            'assoc_type': ('django.db.models.fields.TextField', [], {'max_length': '64'}),
            'handle': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'issued': ('django.db.models.fields.IntegerField', [], {}),
            'lifetime': ('django.db.models.fields.IntegerField', [], {}),
            'secret': ('django.db.models.fields.TextField', [], {'max_length': '255'}),
            'server_url': ('django.db.models.fields.TextField', [], {'max_length': '2047'})
        },
        u'django_openid_auth.nonce': {
            'Meta': {'object_name': 'Nonce'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'salt': ('django.db.models.fields.CharField', [], {'max_length': '40'}),
            'server_url': ('django.db.models.fields.CharField', [], {'max_length': '2047'}),
            'timestamp': ('django.db.models.fields.IntegerField', [], {})
        },
        u'django_openid_auth.useropenid': {
            'Meta': {'object_name': 'UserOpenID'},
            'claimed_id': ('django.db.models.fields.TextField', [], {'unique': 'True', 'max_length': '2047'}),
            'display_id': ('django.db.models.fields.TextField', [], {'max_length': '2047'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']"})
        }
    }

    complete_apps = ['django_openid_auth']


# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2008-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Glue between OpenID and django.contrib.auth."""

from __future__ import unicode_literals

import re

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.core.exceptions import ImproperlyConfigured
from openid.consumer.consumer import SUCCESS
from openid.extensions import ax, sreg, pape

from django_openid_auth import teams
from django_openid_auth.models import UserOpenID
from django_openid_auth.exceptions import (
    IdentityAlreadyClaimed,
    DuplicateUsernameViolation,
    MissingUsernameViolation,
    MissingPhysicalMultiFactor,
    RequiredAttributeNotReturned,
)
from django_openid_auth.signals import openid_duplicate_username


User = get_user_model()


def get_user_group_model():
    """Returns the model used for mapping users to groups."""
    user_group_model_name = getattr(settings, 'AUTH_USER_GROUP_MODEL', None)
    if user_group_model_name is None:
        return User.groups.through
    else:
        try:
            # django.apps available starting from django 1.7
            from django.apps import apps
            get_model = apps.get_model
            args = (user_group_model_name,)
        except ImportError:
            # if we can't import, then it must be django 1.6, still using
            # the old django.db.models.loading code
            from django.db.models.loading import get_model
            app_label, model_name = user_group_model_name.split('.', 1)
            args = (app_label, model_name)
        try:
            model = get_model(*args)
            if model is None:
                # in django 1.6 referring to a non-installed app will
                # return None for get_model, but in 1.7 onwards it will
                # raise a LookupError exception.
                raise LookupError()
            return model
        except ValueError:
            raise ImproperlyConfigured(
                "AUTH_USER_GROUP_MODEL must be of the form "
                "'app_label.model_name'")
        except LookupError:
            raise ImproperlyConfigured(
                "AUTH_USER_GROUP_MODEL refers to model '%s' that has not been "
                "installed" % user_group_model_name)


UserGroup = get_user_group_model()


class OpenIDBackend(object):
    """A django.contrib.auth backend that authenticates the user based on
    an OpenID response."""

    supports_object_permissions = False
    supports_anonymous_user = True

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

    def authenticate(self, **kwargs):
        """Authenticate the user based on an OpenID response."""
        # Require that the OpenID response be passed in as a keyword
        # argument, to make sure we don't match the username/password
        # calling conventions of authenticate.

        openid_response = kwargs.get('openid_response')
        if openid_response is None:
            return None

        if openid_response.status != SUCCESS:
            return None

        user = None
        try:
            user_openid = UserOpenID.objects.get(
                claimed_id__exact=openid_response.identity_url)
        except UserOpenID.DoesNotExist:
            if getattr(settings, 'OPENID_CREATE_USERS', False):
                user = self.create_user_from_openid(openid_response)
        else:
            user = user_openid.user

        if user is None:
            return None

        if getattr(settings, 'OPENID_UPDATE_DETAILS_FROM_SREG', False):
            details = self._extract_user_details(openid_response)
            self.update_user_details(user, details, openid_response)

        if getattr(settings, 'OPENID_PHYSICAL_MULTIFACTOR_REQUIRED', False):
            pape_response = pape.Response.fromSuccessResponse(openid_response)
            key = pape.AUTH_MULTI_FACTOR_PHYSICAL
            if (pape_response is None or
                    key not in pape_response.auth_policies):
                raise MissingPhysicalMultiFactor()

        teams_response = teams.TeamsResponse.fromSuccessResponse(
            openid_response)
        if teams_response:
            self.update_groups_from_teams(user, teams_response)
            self.update_staff_status_from_teams(user, teams_response)

        teams_required = getattr(settings,
                                 'OPENID_LAUNCHPAD_TEAMS_REQUIRED', [])
        if teams_required:
            teams_mapping = self.get_teams_mapping()
            groups_required = [group for team, group in teams_mapping.items()
                               if team in teams_required]
            user_groups = UserGroup.objects.filter(user=user)
            matches = set(groups_required).intersection(
                user_groups.values_list('group__name', flat=True))
            if not matches:
                name = 'OPENID_EMAIL_WHITELIST_REGEXP_LIST'
                whitelist_regexp_list = getattr(settings, name, [])
                for pattern in whitelist_regexp_list:
                    if re.match(pattern, user.email):
                        return user
                return None

        return user

    def _extract_user_details(self, openid_response):
        email = fullname = first_name = last_name = nickname = None
        verified = 'no'
        sreg_response = sreg.SRegResponse.fromSuccessResponse(openid_response)
        if sreg_response:
            email = sreg_response.get('email')
            fullname = sreg_response.get('fullname')
            nickname = sreg_response.get('nickname')
        # If any attributes are provided via Attribute Exchange, use
        # them in preference.
        fetch_response = ax.FetchResponse.fromSuccessResponse(openid_response)
        if fetch_response:
            # The myOpenID provider advertises AX support, but uses
            # attribute names from an obsolete draft of the
            # specification.  We check for them first so the common
            # names take precedence.
            email = fetch_response.getSingle(
                'http://schema.openid.net/contact/email', email)
            fullname = fetch_response.getSingle(
                'http://schema.openid.net/namePerson', fullname)
            nickname = fetch_response.getSingle(
                'http://schema.openid.net/namePerson/friendly', nickname)

            email = fetch_response.getSingle(
                'http://axschema.org/contact/email', email)
            fullname = fetch_response.getSingle(
                'http://axschema.org/namePerson', fullname)
            first_name = fetch_response.getSingle(
                'http://axschema.org/namePerson/first', first_name)
            last_name = fetch_response.getSingle(
                'http://axschema.org/namePerson/last', last_name)
            nickname = fetch_response.getSingle(
                'http://axschema.org/namePerson/friendly', nickname)
            verified = fetch_response.getSingle(
                'http://ns.login.ubuntu.com/2013/validation/account', verified)

        if fullname and not (first_name or last_name):
            # Django wants to store first and last names separately,
            # so we do our best to split the full name.
            fullname = fullname.strip()
            split_names = fullname.rsplit(None, 1)
            if len(split_names) == 2:
                first_name, last_name = split_names
            else:
                first_name = ''
                last_name = fullname

        verification_scheme_map = getattr(
            settings, 'OPENID_VALID_VERIFICATION_SCHEMES', {})
        valid_schemes = verification_scheme_map.get(
            openid_response.endpoint.server_url,
            verification_scheme_map.get(None, ()))
        verified = (verified in valid_schemes)

        return dict(email=email, nickname=nickname, account_verified=verified,
                    first_name=first_name, last_name=last_name)

    def _get_preferred_username(self, nickname, email):
        if nickname:
            return nickname
        if email and getattr(settings, 'OPENID_USE_EMAIL_FOR_USERNAME', False):
            suggestion = ''.join([x for x in email if x.isalnum()])
            if suggestion:
                return suggestion
        return 'openiduser'

    def _get_available_username_for_nickname(self, nickname, identity_url):
        # If we don't have a nickname, and we're not being strict, use a
        # default
        nickname = nickname or 'openiduser'

        # See if we already have this nickname assigned to a username
        if not User.objects.filter(username=nickname).exists():
            return nickname

        # Check if we already have nickname+i for this identity_url
        try:
            user_openid = UserOpenID.objects.get(
                claimed_id=identity_url,
                user__username__startswith=nickname)
            # No exception means we have an existing user for this identity
            # that starts with this nickname.

            # If they are an exact match, the user already exists and hasn't
            # changed their username, so continue to use it
            if nickname == user_openid.user.username:
                return nickname

            # It is possible we've had to assign them to nickname+i already.
            oid_username = user_openid.user.username
            if len(oid_username) > len(nickname):
                try:
                    # check that it ends with a number
                    int(oid_username[len(nickname):])
                    return oid_username
                except ValueError:
                    # username starts with nickname, but isn't nickname+#
                    pass
        except UserOpenID.DoesNotExist:
            # No user associated with this identity_url
            pass

        # Pick a username for the user based on their nickname,
        # checking for conflicts.  Start with number of existing users who's
        # username starts with this nickname to avoid having to iterate over
        # all of the existing ones.
        i = User.objects.filter(username__startswith=nickname).count() + 1
        username = nickname
        while User.objects.filter(username=username).exists():
            username = nickname + str(i)
            i += 1

        return username

    def _ensure_available_username(self, nickname, identity_url):
        if not nickname:
            raise MissingUsernameViolation()

        # As long as the `QuerySet` does not get evaluated, no
        # caching should be involved in our multiple `exists()`
        # calls. See docs for details: http://bit.ly/2aYCmkw
        user_with_same_username = User.objects.exclude(
            useropenid__claimed_id=identity_url
        ).filter(username=nickname)

        if user_with_same_username.exists():
            # Notify any listeners that a duplicated username was
            # found and give the opportunity to handle conflict.
            openid_duplicate_username.send(sender=User, username=nickname)

            # Check for conflicts again as the signal could have handled it.
            if user_with_same_username.exists():
                raise DuplicateUsernameViolation(
                    "The username (%s) with which you tried to log in is "
                    "already in use for a different account." % nickname)

    def _get_available_username(self, nickname, identity_url):
        if getattr(settings, 'OPENID_STRICT_USERNAMES', False):
            self._ensure_available_username(nickname, identity_url)
        else:
            nickname = self._get_available_username_for_nickname(
                nickname, identity_url)
        return nickname

    def create_user_from_openid(self, openid_response):
        details = self._extract_user_details(openid_response)
        required_attrs = getattr(settings, 'OPENID_SREG_REQUIRED_FIELDS', [])
        if getattr(settings, 'OPENID_STRICT_USERNAMES', False):
            required_attrs.append('nickname')

        for required_attr in required_attrs:
            if required_attr not in details or not details[required_attr]:
                raise RequiredAttributeNotReturned(
                    "An attribute required for logging in was not "
                    "returned ({0}).".format(required_attr))

        nickname = self._get_preferred_username(
            details['nickname'], details['email'])
        email = details['email'] or ''

        username = self._get_available_username(
            nickname, openid_response.identity_url)

        user = User.objects.create_user(username, email, password=None)
        self.associate_openid(user, openid_response)
        self.update_user_details(user, details, openid_response)

        return user

    def associate_openid(self, user, openid_response):
        """Associate an OpenID with a user account."""
        # Check to see if this OpenID has already been claimed.
        try:
            user_openid = UserOpenID.objects.get(
                claimed_id__exact=openid_response.identity_url)
        except UserOpenID.DoesNotExist:
            user_openid = UserOpenID(
                user=user,
                claimed_id=openid_response.identity_url,
                display_id=openid_response.endpoint.getDisplayIdentifier())
            user_openid.save()
        else:
            if user_openid.user != user:
                raise IdentityAlreadyClaimed(
                    "The identity %s has already been claimed"
                    % openid_response.identity_url)

        return user_openid

    def update_user_details(self, user, details, openid_response):
        updated = False
        if details['first_name']:
            user.first_name = details['first_name'][:30]
            updated = True
        if details['last_name']:
            user.last_name = details['last_name'][:30]
            updated = True
        if details['email']:
            user.email = details['email']
            updated = True
        if getattr(settings, 'OPENID_FOLLOW_RENAMES', False):
            user.username = self._get_available_username(
                details['nickname'], openid_response.identity_url)
            updated = True
        account_verified = details.get('account_verified', None)
        if (account_verified is not None):
            permission = Permission.objects.get(codename='account_verified')
            perm_label = '%s.%s' % (permission.content_type.app_label,
                                    permission.codename)
            if account_verified and not user.has_perm(perm_label):
                user.user_permissions.add(permission)
            elif not account_verified and user.has_perm(perm_label):
                user.user_permissions.remove(permission)

        if updated:
            user.save()

    def get_teams_mapping(self):
        teams_mapping_auto = getattr(
            settings, 'OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO', False)
        teams_mapping_auto_blacklist = getattr(
            settings, 'OPENID_LAUNCHPAD_TEAMS_MAPPING_AUTO_BLACKLIST', [])
        teams_mapping = getattr(settings, 'OPENID_LAUNCHPAD_TEAMS_MAPPING', {})
        if teams_mapping_auto:
            # ignore teams_mapping. use all django-groups
            teams_mapping = dict()
            all_groups = Group.objects.exclude(
                name__in=teams_mapping_auto_blacklist)
            for group in all_groups:
                teams_mapping[group.name] = group.name
        return teams_mapping

    def update_groups_from_teams(self, user, teams_response):
        teams_mapping = self.get_teams_mapping()
        if len(teams_mapping) == 0:
            return

        mapping = [
            teams_mapping[lp_team] for lp_team in teams_response.is_member
            if lp_team in teams_mapping]
        user_groups = UserGroup.objects.filter(user=user)
        matching_groups = user_groups.filter(
            group__name__in=teams_mapping.values())
        current_groups = set(
            user_group.group for user_group in matching_groups)
        desired_groups = set(Group.objects.filter(name__in=mapping))
        groups_to_remove = current_groups - desired_groups
        groups_to_add = desired_groups - current_groups
        user_groups.filter(group__in=groups_to_remove).delete()
        for group in groups_to_add:
            UserGroup.objects.create(user=user, group=group)

    def update_staff_status_from_teams(self, user, teams_response):
        if not hasattr(settings, 'OPENID_LAUNCHPAD_STAFF_TEAMS'):
            return

        staff_teams = getattr(settings, 'OPENID_LAUNCHPAD_STAFF_TEAMS', [])
        user.is_staff = False

        for lp_team in teams_response.is_member:
            if lp_team in staff_teams:
                user.is_staff = True
                break

        user.save()
